package com.laguerre.kan;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;

import android.opengl.GLES20;

public final class ImageEffects {
    private static final String vShaderCode =
    		"precision lowp float;" +
    		"attribute vec4 position;" +
    		"attribute vec2 uv;" +
    		"varying vec2 texCoord;" +
    		"void main() {" +
    		"  gl_Position = position;" +
    		"  texCoord = uv;" +
    		"}";

    public static String getBasicVertexShaderCode() {
    	return vShaderCode;
    }

    private static final String fShaderCode =
    		"precision lowp float;" +
    		"varying vec2 texCoord;" +
    		"uniform sampler2D prevRenderedTex;" +
    		"void main() {" +
    		"  gl_FragColor = texture2D(prevRenderedTex, texCoord);" +
    		"}";

    private static final float[] quadVertices = {
    	-1.0f,  1.0f, 0.0f, 1.0f,   // top left
    	-1.0f, -1.0f, 0.0f, 1.0f,   // bottom left
    	 1.0f, -1.0f, 0.0f, 1.0f,   // bottom right
    	 1.0f,  1.0f, 0.0f, 1.0f    // top right
};

    private static final float[] quadUV = {
    	0.0f, 1.0f, // top left
    	0.0f, 0.0f, // bottom left
    	1.0f, 0.0f, // bottom right
    	1.0f, 1.0f  // top right
    };

    private static short[] drawList = {
    	0, 1, 2, 0, 2, 3
    };

    private static final int POSITION_DATA_SIZE = 4;
    private static final int UV_DATA_SIZE = 2;

    public static enum FinalDestination {
    	SCREEN,
    	TEXTURE
    }

    /**
     * Tracks that init() has been called.
     */
    private boolean initialized = false;
    /**
     * The {@link RenderTexture} this class will render to. It <b>must</b> be set
     * by a call to <code>setRenderTexture</code>.
     */
    private RenderTexture texToRender;
    /**
     * Whether the last pass will render to screen or to <code>texToRender</code>.
     */
    private FinalDestination renderingMode;
    /**
     * The handler of the frame buffer created by <code>setupFBO</code>.
     */
    private int fbHandle = -1;
    /**
     * Tracks that OpenGL parameters have been set for the current FBO.
     * The method <code>setupFBO</code> will set this boolean to <code>false</code>
     * and <code>preRender</code> will turn it to <code>true</code> once parameters
     * have been passed (I.e. after binding to the FBO).
     */
    private boolean glFBOParamsInitialized;
    /**
     * Similar to <code>glFBOParamsInitialized</code> but for the frame buffer
     * numbered 0 (i.e. the screen).
     */
    private boolean glScreenParamsInitialized = false;
    private FloatBuffer packedDataBuffer;
	private ShortBuffer drawListBuffer;
    private final Shader defaultShader = new Shader();
    private Shader currentShader = null;
    private int passNum;
    private boolean needsPass;

	ImageEffects() {
	}

	/**
	 * Builds the geometry of the quad that will support the render texture.
	 * This method <b>must</b> be called after the constructor, prior to any
	 * operation.
	 */
	public void init() {
		final int vertexCount = quadVertices.length / POSITION_DATA_SIZE;
		final int dataSize = POSITION_DATA_SIZE + UV_DATA_SIZE;

    	final float[] vertexData = new float[vertexCount * dataSize];
    	for (int i = 0; i < vertexCount; i++) {
    		final int start = i * dataSize;
    		// Position
    		final int readPositionStart = POSITION_DATA_SIZE * i;
    		vertexData[start    ] = quadVertices[readPositionStart    ];  // x
    		vertexData[start + 1] = quadVertices[readPositionStart + 1];  // y
    		vertexData[start + 2] = quadVertices[readPositionStart + 2];  // z
    		vertexData[start + 3] = quadVertices[readPositionStart + 3];  // w
    		// UV
    		vertexData[start + 4] = quadUV[UV_DATA_SIZE * i    ];  // u
    		vertexData[start + 5] = quadUV[UV_DATA_SIZE * i + 1];  // v
    	}

    	final ByteBuffer bb = ByteBuffer.allocateDirect(vertexCount * dataSize * Float.SIZE / 8);
    	bb.order(ByteOrder.nativeOrder());
    	this.packedDataBuffer = bb.asFloatBuffer();
    	this.packedDataBuffer.put(vertexData);
    	this.packedDataBuffer.position(0);

    	final ByteBuffer bbDrawList = ByteBuffer.allocateDirect(drawList.length * Short.SIZE / 8);
    	bbDrawList.order(ByteOrder.nativeOrder());
    	this.drawListBuffer = bbDrawList.asShortBuffer();
    	this.drawListBuffer.put(drawList);
    	this.drawListBuffer.position(0);

		final PassShader passShader = this.defaultShader.addPassShader(0);
		passShader.createProgram();
		passShader.attachVertexShader(vShaderCode);
		passShader.attachFragmentShader(fShaderCode);
		passShader.linkProgram();

		// Use the default shader.
		resetToDefaultShader();

		this.initialized = true;
		this.glFBOParamsInitialized = false;

		setRenderingMode(ImageEffects.FinalDestination.TEXTURE);
	}

	/**
	 * Tells whether the method <code>init</code> has been called.
	 * @return <code>true</code> if it was called, <code>false</code> otherwise.
	 */
	public boolean isInitialized() {
		return this.initialized;
	}

	/**
	 * Sets the {@link RenderTexture} to which this class will render to.
	 * Note that this method sets up also an FBO at that time, so you <b>must</b>
	 * call it right after <code>init</code> and prior to any operation.
	 * @param renderTexture
	 */
    public void setRenderTexture(RenderTexture renderTexture) {
    	if (renderTexture != this.texToRender) {
    		this.texToRender = renderTexture;
    		setupFBO();
    	}
    }

	private void setupFBO() {
		// Deletes the previous FBO, if any.
		freeResources();

		final int[] names = { -1 };
		// FB
		GLES20.glGenFramebuffers(1, names, 0);
		this.fbHandle = names[0];
		this.glFBOParamsInitialized = false;
	}

	/**
	 * Free resources: the frame buffer is deleted.
	 */
	public void freeResources() {
		final int[] rendBuff = new int[] { this.fbHandle };
		if (this.fbHandle > 0) {
			GLES20.glDeleteFramebuffers(1, rendBuff, 0);
			this.fbHandle = -1;
		}
	}

	public void resetToDefaultShader() {
		this.currentShader = this.defaultShader;
	}

	public void useAlternativeShader(Shader shader) {
		this.currentShader = shader;
	}

    public RenderTexture getSharedRenderTexture() {
    	return this.texToRender;
    }

    public void setRenderingMode(FinalDestination renderingMode) {
    	this.renderingMode = renderingMode;
    }

    public FinalDestination getRenderingMode() {
    	return this.renderingMode;
    }

    public void setToFirstPass() {
    	this.passNum = 0;
    	this.needsPass = true;
    }

	public int getCurrentPass() {
		return this.passNum;
	}

    public boolean renderOnePass() {
    	if (!this.needsPass) {
    		return false;
    	}

    	preRender();

    	final int byteStride = (POSITION_DATA_SIZE + UV_DATA_SIZE) * Float.SIZE / 8;
    	this.currentShader.setShaderBuffer(
    			"position", POSITION_DATA_SIZE, byteStride, this.packedDataBuffer, 0);
    	this.currentShader.setShaderBuffer(
    			"uv", UV_DATA_SIZE, byteStride, this.packedDataBuffer, POSITION_DATA_SIZE);

    	this.currentShader.setShaderTexture("prevRenderedTex", this.texToRender.getPrevTextureHandle());

    	this.currentShader.setPass(this.passNum);

    	this.currentShader.use();

    	GLES20.glDrawElements(GLES20.GL_TRIANGLES, drawList.length,
                GLES20.GL_UNSIGNED_SHORT, this.drawListBuffer);

    	this.currentShader.releaseAllShaderBuffers();

    	this.passNum++;
    	this.needsPass = (this.passNum < this.currentShader.passCount());

    	// Avoids to swap buffers after the last pass.
    	if (this.needsPass) {
    		GLES20.glBindFramebuffer(GLES20.GL_FRAMEBUFFER, 0);
    		this.texToRender.swapPrevAndCurrentTexture();
    	}

    	return this.needsPass;
    }

	public boolean stillNeedsPass() {
		return this.needsPass;
	}

	private void preRender() {
    	boolean finalPass = (this.passNum + 1 == this.currentShader.passCount());

    	Texture.resetTextureCounter();

    	if (finalPass && this.renderingMode == FinalDestination.SCREEN) {
    		// Rendering to the screen.
    		if (!glScreenParamsInitialized) {
    			initGLParams();
    			glScreenParamsInitialized = true;
    		}
    	} else {
    		// Rendering to a texture.
    		if (this.fbHandle < 1) {
    			throw new Utils.WrongFrameBufferHandleException(
    					"FBO not initialized? (fbHandle = " + this.fbHandle + ")");
    		}

    		GLES20.glBindFramebuffer(GLES20.GL_FRAMEBUFFER, this.fbHandle);
    		GLES20.glFramebufferTexture2D(
    				GLES20.GL_FRAMEBUFFER, GLES20.GL_COLOR_ATTACHMENT0,
                    GLES20.GL_TEXTURE_2D, this.texToRender.getTextureHandle(), 0);

    		final int fbStatus = GLES20.glCheckFramebufferStatus(GLES20.GL_FRAMEBUFFER);
    		if (fbStatus != GLES20.GL_FRAMEBUFFER_COMPLETE) {
    			throw new Utils.FrameBufferIncompleteException(
    					"Could not bind frame buffer: " + fbStatus);
    		}

    		if (!this.glFBOParamsInitialized) {
    			initGLParams();
    			this.glFBOParamsInitialized = true;
    		}
    	}

    	// No depth-handling is needed.
    	GLES20.glDisable(GLES20.GL_DEPTH_TEST);
    	// Nor back-face culling because there is only one quad presented
    	// face-to-face to the camera/screen.
    	GLES20.glDisable(GLES20.GL_CULL_FACE);

    	GLES20.glViewport(0, 0, this.texToRender.getWidth(), this.texToRender.getHeight());
    	GLES20.glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
    	GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT);
	}

	/**
	 * Sets OpenGL state values for a FBO that was just created (i.e. after a
	 * call to <code>setupFBO</code>. Called by <code>preRender</code>.
	 */
	private void initGLParams() {
		GLES20.glDisable(GLES20.GL_DEPTH_TEST);
		// cull backface
		GLES20.glEnable(GLES20.GL_CULL_FACE);
	}
}
