package com.laguerre.kan;

import java.nio.FloatBuffer;

import android.opengl.GLES20;
import android.util.SparseArray;

public class MeshRenderer extends Renderer {
	private final SparseArray<Material> materialPerSubMesh = new SparseArray<Material>();
	protected int passNum = 0;
	protected boolean needsPass = true;

	public final void setMaterial(Material mat) {
		setMaterial(mat, 0);
	}

	@Override
	public final void setMaterial(Material mat, int subMeshId) {
		this.materialPerSubMesh.put(subMeshId, mat);
	}

	@Override
	public final Material getSharedMaterial(int subMeshId) {
		return this.materialPerSubMesh.get(subMeshId);
	}

	@Override
	public void setToFirstPass() {
		this.passNum = 0;
		this.needsPass = true;
	}

	@Override
	public int getCurrentPass() {
		return this.passNum;
	}

	/**
	 * Shader values that are common to every shader of this {@link SceneObject}.
	 */
	private final Shader.ModelRelativeData shaderValues = new Shader.ModelRelativeData();

	// Used only by renderOnePass() to avoid many memory allocations.
	private final float[] renderBufferMat4 = new float[16];
	private final float[] renderBufferInvMat4 = new float[16];

	@Override
	public boolean renderOnePass(Camera camera) {
		if (!this.needsPass) {
			return false;
		}

		// will be updated once a rendering occurred.
		this.needsPass = false;

		final SceneObject parentSOb = getSharedSOb();
		final Mesh mesh = parentSOb.getAnyComponent(Mesh.class);
		if (mesh != null) {
			final Transform t = parentSOb.getSharedTransform();

			// Model matrix and the inverse
			shaderValues.setModelMatrix(
					t.getSharedWorldMatrix4().getSharedComponents(),
					t.getSharedInverseWorldMatrix4().getSharedComponents());

			// Other shader values that are camera dependent only.
			final Camera.RenderingValues cameraVals = camera.getSharedRenderingValues();

			// View-model matrix and the inverse
			Transform.combine(this.renderBufferMat4, cameraVals.getSharedViewMatrix(), shaderValues.getSharedModelMatrix());
			Transform.combine(this.renderBufferInvMat4, shaderValues.getSharedInvModelMatrix(), cameraVals.getSharedInvViewMatrix());
			shaderValues.setModelViewMatrix(this.renderBufferMat4, this.renderBufferInvMat4);

			// Proj-view-model matrix and the inverse
			Transform.combine(this.renderBufferMat4, cameraVals.getSharedProjMatrix(), shaderValues.getSharedModelViewMatrix());
			Transform.combine(this.renderBufferInvMat4, shaderValues.getSharedInvModelViewMatrix(), cameraVals.getSharedInvProjMatrix());
			shaderValues.setModelViewProjModelMatrix(this.renderBufferMat4, this.renderBufferInvMat4);

			for (int i = 0, iEnd = mesh.getNbOfSubMeshes(); i < iEnd; i++) {
				final Material mat = this.materialPerSubMesh.get(i);
				if (mat != null) {
					// Initialize the texture counter for all the Materials of this sub-mesh.
					Texture.resetTextureCounter();

					if (mat.isPassAvailable(this.passNum)) {
						mat.setShaderVector("viewSize", cameraVals.getSharedViewSize());
						//
						mat.setShaderVector("eyePosition", cameraVals.getSharedEyePosition());
						//
						mat.setShaderMatrix4("ViewMatrix", cameraVals.getSharedViewMatrix());
						mat.setShaderMatrix4("invViewMatrix", cameraVals.getSharedInvViewMatrix());
						//
						mat.setShaderMatrix4("ProjMatrix", cameraVals.getSharedProjMatrix());
						mat.setShaderMatrix4("invProjMatrix", cameraVals.getSharedInvProjMatrix());
						//
						mat.setShaderMatrix4("VPMatrix", cameraVals.getSharedViewProjMatrix());
						mat.setShaderMatrix4("invVPMatrix", cameraVals.getSharedInvViewProjMatrix());
						//
						mat.setShaderMatrix4("ModelMatrix", shaderValues.getSharedModelMatrix());
						mat.setShaderMatrix4("invModelMatrix", shaderValues.getSharedInvModelMatrix());
						//
						mat.setShaderMatrix4("MVMatrix", shaderValues.getSharedModelViewMatrix());
						mat.setShaderMatrix4("invMVMatrix", shaderValues.getSharedInvModelViewMatrix());
						//
						mat.setShaderMatrix4("MVPMatrix", shaderValues.getSharedModelViewProjMatrix());
						mat.setShaderMatrix4("invMVPMatrix", shaderValues.getSharedInvModelViewProjMatri());
						//
						final int byteStride = mesh.getVertexPackedDataSize(i) * Float.SIZE / 8;
						final FloatBuffer packedDataBuffer = mesh.getSharedVertexPackedData(i);
						mat.setShaderBuffer(
								"position", Mesh.POSITION_DATA_SIZE, byteStride,
								packedDataBuffer, 0);
						mat.setShaderBuffer(
								"normal", Mesh.NORMAL_DATA_SIZE, byteStride,
								packedDataBuffer, Mesh.POSITION_DATA_SIZE);
						mat.setShaderBuffer(
								"uv", Mesh.UV_DATA_SIZE, byteStride,
								packedDataBuffer, Mesh.POSITION_DATA_SIZE + Mesh.NORMAL_DATA_SIZE);

						mat.setShaderValue(
								"hasRenderedTex",
								camera.getSharedRenderTexture().hasSwapped());
						mat.setShaderTexture(
								"prevRenderedTex",
								camera.getSharedRenderTexture().getPrevTextureHandle());

						mat.setShaderVector("ambientLight",
								RenderingSettings.getInstance().getSharedAmbientLight());

						Lights.getInstance().passLightsToShader(camera, mat.getSharedShader());

						mat.setPass(this.passNum);

						mat.use();

						// Still passes required?
						final int nbPasses = mat.passCount();
						// 1st pass = 0!
						if (this.passNum + 1 < nbPasses) {
							this.needsPass = true;
						}
					}
				}

		    	GLES20.glDrawElements(
		    			GLES20.GL_TRIANGLES, mesh.getDrawListCount(i),
		                GLES20.GL_UNSIGNED_SHORT, mesh.getSharedDrawListBuffer(i));

				if (mat != null) {
					mat.releaseAllShaderBuffers();
				}
			}
		}

		// Prepare for the next run.
		this.passNum++;

		return this.needsPass;
	}

	@Override
	public boolean stillNeedsPass() {
		return this.needsPass;
	}
}
