package com.laguerre.kan;

import java.util.ArrayList;
import java.util.List;

public final class ParticleSystem extends Component {
	public static final class Particle {
		private final Vector3 position = Vector3.getNewZeroVec();
		private final Vector3 velocity = Vector3.getNewZeroVec();
		/**
		 * Time in seconds this particle should exist. Below 0, it will be ignored.
		 */
		private float lifeTime = -1f;
		private float size = 0.2f;
		private float rotation = 0f;
		private float angularVelocity = 0f;
		private final Color color = Color.getNewWhite();

		public Particle() {
		}

		public void setPosition(Vector3 p) {
			this.position.setValues(p);
		}

		public void setPosition(float x, float y, float z) {
			this.position.setValues(x, y, z);
		}

		public Vector3 getPosition() {
			return new Vector3(this.position);
		}

		public Vector3 getPosition(Vector3 result) {
			result.setValues(this.position);
			return result;
		}

		/**
		 * Gets a direct reference to the inner <code>Vector3</code> that contains
		 * position in order to avoid memory allocations. Please, do <b>not</b>
		 * update position components by this way, use <code>setPosition</code>.
		 * @return A reference to the position.
		 */
		public Vector3 getSharedPosition() {
			return this.position;
		}

		public void setVelocity(Vector3 v) {
			this.velocity.setValues(v);
		}

		public void setVelocity(float x, float y, float z) {
			this.velocity.setValues(x, y, z);
		}

		public Vector3 getVelocity() {
			return new Vector3(this.velocity);
		}

		public Vector3 getVelocity(Vector3 result) {
			result.setValues(this.velocity);
			return result;
		}

		/**
		 * Gets a direct reference to the inner <code>Vector3</code> that contains
		 * velocity in order to avoid memory allocations. Please, do <b>not</b>
		 * update velocity components by this way, use <code>setVelocity</code>.
		 * @return A reference to the velocity.
		 */
		public Vector3 getSharedVelocity() {
			return this.velocity;
		}

		public void setLifetime(float lifetime) {
			this.lifeTime = lifetime;
		}

		public float getLifetime() {
			return this.lifeTime;
		}

		public void setSize(float size) {
			this.size = size;
		}

		public float getSize() {
			return this.size;
		}

		public void setRotation(float rotation) {
			this.rotation = rotation;
		}

		public float getRotation() {
			return this.rotation;
		}

		public void setAngularVelocity(float angularVelocity) {
			this.angularVelocity = angularVelocity;
		}

		public float getAngularVelocity() {
			return this.angularVelocity;
		}

		public void setColor(Color color) {
			this.color.setValues(color);
		}

		public void setColor(float r, float g, float b, float a) {
			this.color.setValues(r, g, b, a);
		}

		public Color getColor() {
			return new Color(this.color);
		}

		public Color getColor(Color result) {
			result.setValues(this.color);
			return result;
		}

		/**
		 * Gets a direct reference to the inner <code>Color</code> that contains
		 * color in order to avoid memory allocations. Please, do <b>not</b>
		 * update color components by this way, use <code>setColor</code>.
		 * @return A reference to the velocity.
		 */
		public Color getSharedColor() {
			return this.color;
		}
	}

	public static interface IParticleVisitor {
		public void visitParticle(ParticleSystem.Particle particle);
	}

	public static interface EmitterShape {
		public void setParameters(ParticleSystem particleSystem, Particle particle);
	}

	public static interface IParticleModule {
		public void updateParticle(ParticleSystem particleSystem, ParticleSystem.Particle particle);
	}

	private float startLifeTime = 1f;
	private float startSize = 0.2f;
	private float startRotation = 0.0f;
	private float startSpeed = 0.3f;
	private final Color startColor = Color.getNewWhite();
	private int maxParticleCount = 0;
	private boolean playing = false;
	/**
	 * Seconds elapsed since the last time particles have been emitted automatically
	 * because of the "play" mode.
	 */
	private float timeSincePrevPlayingEmit = 0f;
	private final ArrayList<ParticleSystem.Particle> particles = new ArrayList<ParticleSystem.Particle>();
	private EmitterShape emitterShape;
	/**
	 * Number of particles to emit per seconds, if playing.
	 */
	private float rateEmission = 0f;
	private final ArrayList<Burst> bursts = new ArrayList<ParticleSystem.Burst>();
	private final ArrayList<ParticleSystem.IParticleModule> particleModules = new ArrayList<ParticleSystem.IParticleModule>();

	private int nbAliveParticles = 0;

	ParticleSystem() {
		setMaxParticleCount(100);
	}

	public void setStartLifeTime(float startLifeTime) {
		this.startLifeTime = startLifeTime;
	}

	public float getStartLifeTime() {
		return this.startLifeTime;
	}

	public void setStartSize(float startSize) {
		this.startSize = startSize;
	}

	public float getStartSize() {
		return this.startSize;
	}

	public void setStartRotation(float startRotation) {
		this.startRotation = startRotation;
	}

	public float getStartRotation() {
		return this.startRotation;
	}

	public void setStartSpeed(float startSpeed) {
		this.startSpeed = startSpeed;
	}

	public float getStartSpeed() {
		return this.startSpeed;
	}

	public void setStartColor(float r, float g, float b, float a) {
		this.startColor.setValues(r, g, b, a);
	}

	public void setStartColor(Color color) {
		this.startColor.setValues(color);
	}

	public Color getStartColor(Color result) {
		result.setValues(this.startColor);
		return result;
	}

	public Color getStartColor() {
		return new Color(this.startColor);
	}

	public Color getSharedStartColor() {
		return this.startColor;
	}

	public void setMaxParticleCount(int maxParticleCount) {
		this.maxParticleCount = maxParticleCount;
		updateParticlePopulation();
	}

	public int getMaxParticleCount() {
		return this.maxParticleCount;
	}

	/**
	 * Gets the number of alive particles. This number is updated by calls to
	 * <code>doLifeStep</code>.
	 * @return the number of living particles.
	 */
	public int countAliveParticles() {
		return this.nbAliveParticles;
	}

	/**
	 * According to <code>maxParticleCount</code>, instantiates {@link Particle}
	 * objects to fill <code>particles</code> until <code>maxParticleCount</code>
	 * is reached, or removes {@link Particle} objects created previously from
	 * <code>particles</code>.
	 */
	private void updateParticlePopulation() {
		final int prevSize = this.particles.size();
		if (prevSize < this.maxParticleCount) {
			this.particles.ensureCapacity(this.maxParticleCount);
			for (int i = prevSize; i < this.maxParticleCount; i++) {
				this.particles.add(new Particle());
			}
		} else if (prevSize > this.maxParticleCount) {
			if (this.maxParticleCount > 0) {
				final List<Particle> keptParticles = this.particles.subList(0, this.maxParticleCount);
				this.particles.clear();
				this.particles.addAll(keptParticles);
			} else {
				this.particles.clear();
			}
		}
	}

	public void setRateEmission(float rateEmission) {
		this.rateEmission = rateEmission;
	}

	public float getRateEmission() {
		return this.rateEmission;
	}

	public void setEmitterShape(EmitterShape emitterShape) {
		this.emitterShape = emitterShape;
	}

	public void addParticleModule(IParticleModule particleModule) {
		this.particleModules.add(particleModule);
	}

	public void removeParticleModule(IParticleModule particleModule) {
		this.particleModules.remove(particleModule);
	}

	public int particleModuleCount() {
		return this.particleModules.size();
	}

	public void getParticleModules(ArrayList<ParticleSystem.IParticleModule> result) {
		result.addAll(this.particleModules);
	}

	public void getParticleModules(ParticleSystem.IParticleModule[] result) {
		this.particleModules.toArray(result);
	}

	public void visitEachAliveParticle(IParticleVisitor visitor) {
		for (int i = 0, iEnd = this.particles.size(); i < iEnd; i++) {
			final Particle p = this.particles.get(i);
			if (p.lifeTime > 0f) {
				visitor.visitParticle(p);
			}
		}
	}

	public void emit(int count) {
		final int nbParticles = Math.min(this.particles.size(), count);
		int pIndex = 0;
		final int particlesSize = this.particles.size();
		// Ignore particles in use...
		while (pIndex < particlesSize && this.particles.get(pIndex).lifeTime >= 0) {
			pIndex++;
		}
		// Here, we may find particles to emit.
		for (int pEmittedNb = 0; pEmittedNb < nbParticles && pIndex < particlesSize; pIndex++) {
			final Particle p = this.particles.get(pIndex);
			if (p.lifeTime < 0) {
				if (this.emitterShape != null) {
					this.emitterShape.setParameters(this, p);
				}
				// Puts the particle at the right place, with following rotation.
				getSharedTransform().transformWorldPoint(p.position, p.position);
				// Orient correctly velocity.
				getSharedTransform().transformWorldDirection(p.velocity, p.velocity);
				p.velocity.normalizeSelf();
				p.velocity.mulSelf(this.startSpeed);
				p.lifeTime = this.startLifeTime;
				p.size = this.startSize;
				p.rotation = this.startRotation;
				p.color.setValues(this.startColor);
				p.angularVelocity = 0f;
				pEmittedNb++;
			}
		}
	}

	/**
	 * Time elapsed to render the previous frame. Set by <code>doLifeStep</code>.
	 */
	private float frameTime;
	/**
	 * Value 0.5 * gravity * <code>frameTime</code>^2. Set by <code>doLifeStep</code>.
	 */
	private final Vector3 halfGTimeSquare = Vector3.getNewZeroVec();

	/**
	 * Gets the value of frame time being used by this particle renderer to
	 * synchronize all the particles. This value comes from
	 * <code>Time.getInstance().getLastFrameTime()</code>.
	 * @return the frame time being used by this particle renderer.
	 */
	public float getFrameTime() {
		return this.frameTime;
	}

	public Vector3 getSharedHalfGTimeSquare() {
		return this.halfGTimeSquare;
	}

	/**
	 * Updates the state of the particles.
	 */
	void doLifeStep() {
		// time
		this.frameTime = Time.getInstance().getLastFrameTime();
		// Updates gravity 0.5 * g * t²
		this.halfGTimeSquare.setValues(RenderingSettings.getInstance().getSharedGravity());
		this.halfGTimeSquare.mulSelf(0.5f * frameTime * frameTime);

		// Emit new particles if playing.
		if (this.playing) {
			this.timeSincePrevPlayingEmit += this.frameTime;
			final int amountParticles = (int) (this.timeSincePrevPlayingEmit * this.rateEmission);
			// At each time doLifeStep is called, the amount of particles may
			// equal to 0 because of a low rate emission (e.g. around 10/s), so
			// we may need to wait a little bit more.
			if (amountParticles > 0) {
				emit(amountParticles);
				this.timeSincePrevPlayingEmit = 0f;
			}
			// Burst?
			for (int iBurst = 0, iBurstEnd = this.bursts.size(); iBurst < iBurstEnd; iBurst++) {
				final Burst burst = this.bursts.get(iBurst);
				burst.addElapsedTime(this.frameTime);
				burst.emitIfPossible(this);
			}
		}

		this.nbAliveParticles = 0;
		for (int iParticle = 0, iParticleEnd = this.particles.size(); iParticle < iParticleEnd; iParticle++) {
			final Particle p = this.particles.get(iParticle);
			if (p.lifeTime >= 0) {
				this.nbAliveParticles++;
				for (int iModule = 0, iModuleEnd = this.particleModules.size(); iModule < iModuleEnd; iModule++) {
					this.particleModules.get(iModule).updateParticle(this, p);
				}
				p.lifeTime -= this.frameTime;
			}
		}
	}

	public void play() {
		this.playing = true;
		// Sets the following value to get it > 1 second in doLifeStep and hence
		// and immediate emission.
		this.timeSincePrevPlayingEmit = 1f;
	}

	public void stop() {
		this.playing = false;
	}

	public boolean isPlaying() {
		return this.playing;
	}

	public void addBurst(Burst burst) {
		this.bursts.add(burst);
	}

	public void removeBurst(Burst burst) {
		this.bursts.remove(burst);
	}

	public static final class Burst {
		private float timeAtBurst = 0f;
		private float elapsedTime = 0f;
		private int amount = 0;

		public void setTimeAtBurst(float timeAtBurst) {
			this.timeAtBurst = timeAtBurst;
		}

		public float getTimeAtBurst() {
			return this.timeAtBurst;
		}

		public void setAmount(int amount) {
			this.amount = amount;
		}

		public int getAmount() {
			return this.amount;
		}

		void addElapsedTime(float elapsedTime) {
			this.elapsedTime += elapsedTime;
		}

		void emitIfPossible(ParticleSystem particleSystem) {
			if (this.elapsedTime > this.timeAtBurst) {
				particleSystem.emit(this.amount);
				this.elapsedTime = 0f;
			}
		}
	}

	/**
	 * Emits {@link Particle} objects from a sphere.
	 */
	public static final class SphereShape implements EmitterShape {
		public static enum EmitFrom {
			CENTER,
			SURFACE,
			INSIDE_SPHERE
		}

		private EmitFrom emitFrom = EmitFrom.CENTER;
		private float radius = 1.0f;

		public void setRadius(float radius) {
			this.radius = radius;
		}

		public float getRadius() {
			return this.radius;
		}

		public void setEmitFrom(EmitFrom emitFrom) {
			this.emitFrom = emitFrom;
		}

		public EmitFrom getEmitFrom() {
			return this.emitFrom;
		}

		public void setParameters(ParticleSystem particleSystem, Particle particle) {
			switch (this.emitFrom) {
			case CENTER:
				particle.position.setValues(Vector3.ZERO);
				break;
			case SURFACE:
			case INSIDE_SPHERE:
				setRandomPosition(particle, this.radius, this.emitFrom == EmitFrom.SURFACE);
				break;
			}
			setRandomVelocity(particle, this.radius, this.emitFrom == EmitFrom.SURFACE);
		}

		private void setRandomPosition(Particle p, float maxRadius, boolean onSurface) {
			Utils.randomPointFromSphere(p.position, maxRadius, onSurface);
		}

		private void setRandomVelocity(Particle p, float maxRadius, boolean onSurface) {
			Utils.randomPointFromSphere(p.velocity, maxRadius, onSurface);
		}
	}

	/**
	 * Emits {@link Particle} objects from a cone.
	 */
	public static final class ConeShape implements EmitterShape {
		public static enum EmitFrom {
			ORIGIN,
			SURFACE,
			INSIDE_CONE
		}

		private EmitFrom emitFrom = EmitFrom.ORIGIN;
		private float height = 1.0f;
		private float radius = 1.0f;

		/**
		 * Sets height to <code>1.0</code> and computes <code>radius</code> to
		 * obtain a cone with the provided angle.
		 * @param angleDeg angle in degrees.
		 */
		public void setUnitConeAngle(float angleDeg) {
			this.height = 1.0f;
			this.radius = (float) (Math.sin(Math.toRadians(angleDeg) / 2.0));
		}

		public float getHeight() {
			return height;
		}

		public void setHeight(float height) {
			this.height = height;
		}

		public void setRadius(float radius) {
			this.radius = radius;
		}

		public float getRadius() {
			return this.radius;
		}

		public void setEmitFrom(EmitFrom emitFrom) {
			this.emitFrom = emitFrom;
		}

		public EmitFrom getEmitFrom() {
			return this.emitFrom;
		}

		public void setParameters(ParticleSystem particleSystem, Particle particle) {
			switch (this.emitFrom) {
			case ORIGIN:
				particle.position.setValues(Vector3.ZERO);
				break;
			case SURFACE:
			case INSIDE_CONE:
				setRandomPosition(particle, this.height, this.radius,
						this.emitFrom == EmitFrom.SURFACE);
				break;
			}
			setRandomVelocity(particle, this.radius, this.emitFrom == EmitFrom.SURFACE);
		}

		private void setRandomPosition(Particle p, float maxHeight,
				float maxRadius, boolean onSurface) {
			Utils.randomPointFromCone(p.position, maxHeight, maxRadius, onSurface);
		}

		private void setRandomVelocity(Particle p, float maxRadius, boolean onSurface) {
			Utils.randomPointFromCone(p.velocity, 1.0f, maxRadius, onSurface);
		}
	}

	public static final class ApplyVelocityModule implements IParticleModule {
		private boolean applyGravity = false;

		public void setApplyGravity(boolean applyGravity) {
			this.applyGravity = applyGravity;
		}

		public boolean isApplingGravity() {
			return this.applyGravity;
		}

		@Override
		public void updateParticle(ParticleSystem particleSystem, Particle particle) {
			final float frameTime = particleSystem.getFrameTime();
			particle.position.addSelf(
					particle.velocity.getX() * frameTime,
					particle.velocity.getY() * frameTime,
					particle.velocity.getZ() * frameTime);
			particle.rotation += particle.angularVelocity * frameTime;
			if (this.applyGravity) {
				particle.velocity.addSelf(particleSystem.getSharedHalfGTimeSquare());
			}
		}
	}

	public static final class UpdateColorOverTimeModule implements IParticleModule {
		private final Color initialColor = Color.getNewWhite();
		private final Color finalColor = Color.getNewWhite();

		public void setInitialColor(Color initialColor) {
			this.initialColor.setValues(initialColor);
		}

		public void setInitialColor(float r, float g, float b, float a) {
			this.initialColor.setValues(r, g, b, a);
		}

		public void setFinalColor(Color finalColor) {
			this.finalColor.setValues(finalColor);
		}

		public void setFinalColor(float r, float g, float b, float a) {
			this.finalColor.setValues(r, g, b, a);
		}

		@Override
		public void updateParticle(ParticleSystem particleSystem,
				Particle particle) {
			final float maxLifeTime = particleSystem.getStartLifeTime();
			particle.setColor(
					Utils.mix(this.initialColor.getR(), this.finalColor.getR(), maxLifeTime, 0.0f, particle.lifeTime),
					Utils.mix(this.initialColor.getG(), this.finalColor.getG(), maxLifeTime, 0.0f, particle.lifeTime),
					Utils.mix(this.initialColor.getB(), this.finalColor.getB(), maxLifeTime, 0.0f, particle.lifeTime),
					Utils.mix(this.initialColor.getA(), this.finalColor.getA(), maxLifeTime, 0.0f, particle.lifeTime));
		}
	}
}
