package com.laguerre.mytest;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;

import com.laguerre.kan.ManualMesh;

public class Cube extends ManualMesh {
	private static final float[] cubeVertices = {
		/*
		 *   4+------+ 7      ^ y
		 *   /      / |       |
		 *  0+-----+3 |       0---> x
		 *   |     |  +6     z
		 *   |     | /
		 *  1+-----+2
		 *
		 */

		// front: 0, 1, 2, 0, 2, 3
		//  => 0, 1, 2, 0, 2, 3
		-0.5f,  0.5f,  0.5f, 1.0f,   // 0 - 0
		-0.5f, -0.5f,  0.5f, 1.0f,   // 1 - 1
		 0.5f, -0.5f,  0.5f, 1.0f,   // 2 - 2
		 0.5f,  0.5f,  0.5f, 1.0f,   // 3 - 3

		// right: 3, 2, 6, 3, 6, 7
		//  => 4, 5, 6, 4, 6, 7
		 0.5f,  0.5f,  0.5f, 1.0f,   // 3 - 4
		 0.5f, -0.5f,  0.5f, 1.0f,   // 2 - 5
		 0.5f, -0.5f, -0.5f, 1.0f,   // 6 - 6
		 0.5f,  0.5f, -0.5f, 1.0f,   // 7 - 7

		// back: 7, 6, 5, 7, 5, 4
		//  => 8, 9, 10, 8, 10, 11
		 0.5f,  0.5f, -0.5f, 1.0f,   // 7 - 8
		 0.5f, -0.5f, -0.5f, 1.0f,   // 6 - 9
		-0.5f, -0.5f, -0.5f, 1.0f,   // 5 - 10
		-0.5f,  0.5f, -0.5f, 1.0f,   // 4 - 11

		// left: 4, 5, 1, 4, 1, 0
		//  => 12, 13, 14, 12, 14, 15
		-0.5f,  0.5f, -0.5f, 1.0f,   // 4 - 12
		-0.5f, -0.5f, -0.5f, 1.0f,   // 5 - 13
		-0.5f, -0.5f,  0.5f, 1.0f,   // 1 - 14
		-0.5f,  0.5f,  0.5f, 1.0f,   // 0 - 15

		// top: 4, 0, 3, 4, 3, 7
		//  => 16, 17, 18, 16, 18, 19
		-0.5f,  0.5f, -0.5f, 1.0f,   // 4 - 16
		-0.5f,  0.5f,  0.5f, 1.0f,   // 0 - 17
		 0.5f,  0.5f,  0.5f, 1.0f,   // 3 - 18
		 0.5f,  0.5f, -0.5f, 1.0f,   // 7 - 19

		// bottom: 1, 5, 6, 1, 6, 2
		//  => 20, 21, 22, 20, 22, 23
		-0.5f, -0.5f,  0.5f, 1.0f,   // 1 - 20
		-0.5f, -0.5f, -0.5f, 1.0f,   // 5 - 21
		 0.5f, -0.5f, -0.5f, 1.0f,   // 6 - 22
		 0.5f, -0.5f,  0.5f, 1.0f,   // 2 - 23
	};

	private static final float[] cubeNormals = {
		// front
		 0f,  0f,  1f,    0f,  0f,  1f,    0f,  0f,  1f,    0f,  0f,  1f,
		// right
		 1f,  0f,  0f,    1f,  0f,  0f,    1f,  0f,  0f,    1f,  0f,  0f,
		// back
		 0f,  0f, -1f,    0f,  0f, -1f,    0f,  0f, -1f,    0f,  0f, -1f,
		// left
		-1f,  0f,  0f,   -1f,  0f,  0f,   -1f,  0f,  0f,   -1f,  0f,  0f,
		// top
		 0f,  1f,  0f,    0f,  1f,  0f,    0f,  1f,  0f,    0f,  1f,  0f,
		// bottom
		 0f, -1f,  0f,    0f, -1f,  0f,    0f, -1f,  0f,    0f, -1f,  0f
	};

	private static final float[] cubeUV = {
		/*
		 *  Faces in OpenGL texture:
		 *
		 *  +---+---+---+
		 *  |fr |ri |ba +
		 *  +---+---+---+
		 *  |le |to |bo |
		 *  +---+---+---+
		 *
		 */
		// front
		0.0f, 1.0f,       0.0f, 0.5f,       0.333333f, 0.5f,   0.333333f, 1.0f,
		// right
		0.333333f, 1.0f,  0.333333f, 0.5f,  0.666666f, 0.5f,   0.666666f, 1.0f,
		// back
		0.666666f, 1.0f,  0.666666f, 1.0f,  1.0f, 0.5f,        1.0f, 1.0f,
		// left
		0.0f, 0.5f,       0.0f, 0.0f,       0.333333f, 0.0f,   0.333333f, 0.5f,
		// top
		0.333333f, 0.5f,  0.333333f, 0.0f,  0.666666f, 0.0f,   0.666666f, 0.5f,
		// bottom
		0.666666f, 0.5f,  0.666666f, 0.5f,  1.0f, 0.0f,        1.0f, 0.5f
	};


	private static short[] drawList = {
		// front
		0, 1, 2, 0, 2, 3,
		// right
		4, 5, 6, 4, 6, 7,
		// back
		8, 9, 10, 8, 10, 11,
		// left
		12, 13, 14, 12, 14, 15,
		// top
		16, 17, 18, 16, 18, 19,
		// bottom
		20, 21, 22, 20, 22, 23
	};

	private FloatBuffer packedData;
	private ShortBuffer drawListBuffer;

    @Override
    public int getNbOfSubMeshes() {
    	return 1;
    }

	@Override
	public void build() {
    	final int vertexCount = cubeVertices.length / POSITION_DATA_SIZE;
    	final int dataSize = getVertexPackedDataSize(0);

    	final float[] vertexData = new float[vertexCount * dataSize];
    	for (int i = 0; i < vertexCount; i++) {
    		final int start = i * dataSize;
    		// Position
    		final int readPositionStart = POSITION_DATA_SIZE * i;
    		vertexData[start    ] = cubeVertices[readPositionStart    ];  // x
    		vertexData[start + 1] = cubeVertices[readPositionStart + 1];  // y
    		vertexData[start + 2] = cubeVertices[readPositionStart + 2];  // z
    		vertexData[start + 3] = cubeVertices[readPositionStart + 3];  // w
    		// Normal
    		final int readNormalStart = NORMAL_DATA_SIZE * i;
    		vertexData[start + 4] = cubeNormals[readNormalStart    ];  // x
    		vertexData[start + 5] = cubeNormals[readNormalStart + 1];  // y
    		vertexData[start + 6] = cubeNormals[readNormalStart + 2];  // z
    		// UV
    		vertexData[start + 7] = cubeUV[UV_DATA_SIZE * i    ];  // u
    		vertexData[start + 8] = cubeUV[UV_DATA_SIZE * i + 1];  // v
    	}

    	final ByteBuffer bb = ByteBuffer.allocateDirect(vertexCount * dataSize * Float.SIZE / 8);
    	bb.order(ByteOrder.nativeOrder());
    	this.packedData = bb.asFloatBuffer();
    	this.packedData.put(vertexData);
    	this.packedData.position(0);

    	final ByteBuffer bbDrawList = ByteBuffer.allocateDirect(drawList.length * Short.SIZE / 8);
    	bbDrawList.order(ByteOrder.nativeOrder());
    	this.drawListBuffer = bbDrawList.asShortBuffer();
    	this.drawListBuffer.put(drawList);
    	this.drawListBuffer.position(0);
	}

	@Override
	public FloatBuffer getSharedVertexPackedData(int subMeshId) {
		return this.packedData;
	}

	@Override
	public int getVertexCount(int subMeshId) {
		return cubeVertices.length;
	}

	@Override
	public int getDrawListCount(int subMeshId) {
		return drawList.length;
	}

	@Override
	public ShortBuffer getSharedDrawListBuffer(int subMeshId) {
		return this.drawListBuffer;
	}
}
